typedef struct Rop Rop;
typedef struct Req Req;
typedef struct Trans Trans;

typedef struct Share Share;
typedef struct File File;
typedef struct Find Find;
typedef struct Tree Tree;

struct Rop
{
	int (*strpack)(uchar *, uchar *, uchar *, void *);
	int (*strunpack)(uchar *, uchar *, uchar *, void *);
	int (*namepack)(uchar *, uchar *, uchar *, void *);
	int (*nameunpack)(uchar *, uchar *, uchar *, void *);
	int (*untermstrpack)(uchar *, uchar *, uchar *, void *);
	int (*untermnamepack)(uchar *, uchar *, uchar *, void *);
};

struct Req
{
	int cmd;
	int tid;
	int pid;
	int uid;
	int mid;
	int flags;
	int flags2;

	uchar *lh, *rh, *rp, *re;

	Rop *o;
	char *name;
	void (*respond)(Req *r, int err);
};

struct Trans
{
	int cmd;
	int flags;

	struct {
		struct {
			uchar *b, *p, *e;
		} param, data, setup;
	} in, out;

	Req *r;
	char *name;
	void (*respond)(Trans *t, int err);
};

struct File
{
	int ref;
	int fd;
	int rtype;
	int access;
	int delete;
	char *path;
};

struct Find
{
	int ref;
	int attr;
	char *base;
	char *pattern;
	int index;
	Dir *dotdot;
	Dir *dot;
	Dir *dir;
	int ndir;
};

struct Share
{
	Share *next;

	char *service;
	int stype;

	char *name;
	char *root;
	char *remark;

	char *fsname;
	int namelen;
	vlong allocsize;
	vlong freesize;
	int sectorsize;
	int blocksize;
};

struct Tree
{
	int tid;

	void **file;
	int nfile;

	void **find;
	int nfind;

	Share *share;
};

int debug;
int trspaces;
int needauth;
char *domain;
char *progname;
char *osname;

char *remotesys;
char *remoteuser;
int remotebuffersize;

long starttime;
int tzoff;

enum
{
	BUFFERSIZE = 0x8000,

	/* error class */
	ERRDOS = 1,
	ERRSRV = 2,
	ERRHRD = 3,
	ERRCMD = 0xFF,

	/* error codes */
	ERRbadfunc = 0x1,
	ERRbadfile = 0x2,
	ERRbadpath = 0x3,
	ERRnofids = 0x4,
	ERRnoaccess = 0x5,
	ERRbadfid = 0x6,
	ERRbadmcp = 0x7,
	ERRnomem = 0x8,
	ERRbadmem = 0x9,
	ERRbadenv = 0xA,
	ERRbadformat = 0xB,
	ERRbadaccess = 0xC,
	ERRbaddata = 0xD,
	ERRbaddrive = 0xF,
	ERRremcd = 0x10,
	ERRdiffdevice = 0x11,
	ERRnofiles = 0x12,
	ERRgeneral = 0x1F,
	ERRbadshare = 0x20,
	ERRlock = 0x21,
	ERReof = 0x26,
	ERRunsup = 0x32,
	ERRfilexists = 0x50,
	ERRinvalidparam = 0x57,
	ERRunknownlevel = 0x7C,
	ERRbadpipe = 0xE6,
	ERRinvnetname = 0x06,
	ERRreqnotaccep = 0x47,
	ERRnosuchshare = 0x43,
	ERRerror = 0x1,
	ERRbadpw = 0x2,
	ERRaccess = 0x4,
	ERRinvtid = 0x5,
	ERRinvdevice = 0x7,
	ERRbaduid = 0x5b,

	/* resource type */
	FileTypeDisk = 0,

	/* stype */
	STYPE_DISKTREE = 0,
	STYPE_PRINTQ = 1,
	STYPE_DEVICE = 2,
	STYPE_IPC = 3,

	/* capabilities */
	CAP_UNICODE = 0x4,
	CAP_LARGEFILES = 0x8,
	CAP_NT_SMBS = 0x10,
	CAP_NT_FIND = 0x200,

	/* extended file attributes */
	ATTR_READONLY = 0x1,
	ATTR_HIDDEN = 0x2,
	ATTR_SYSTEM = 0x4,
	ATTR_DIRECTORY = 0x10,
	ATTR_ARCHIVE = 0x20,
	ATTR_NORMAL = 0x80,

	DOSMASK	 = 0x37,

	/* access */
	FILE_READ_DATA = 0x1,
	FILE_WRITE_DATA = 0x2,
	FILE_APPEND_DATA = 0x4,
	FILE_EXECUTE = 0x20,
	FILE_DELETE = 0x10000,
	GENERIC_ALL = 0x10000000,
	GENERIC_EXECUTE = 0x20000000,
	GENERIC_WRITE = 0x40000000,
	GENERIC_READ = 0x80000000,

	READMASK =
		FILE_READ_DATA |
		FILE_EXECUTE |
		GENERIC_ALL |
		GENERIC_EXECUTE |
		GENERIC_READ,

	WRITEMASK = 
		FILE_WRITE_DATA |
		FILE_APPEND_DATA |
		GENERIC_ALL |
		GENERIC_WRITE,

	/* createdisposition */
	FILE_SUPERSEDE = 0,
	FILE_OPEN,
	FILE_CREATE,
	FILE_OPEN_IF,
	FILE_OVERWRITE,
	FILE_OVERWRITE_IF,

	/* createaction */
	FILE_SUPERSEDED = 0,
	FILE_OPEND,
	FILE_CREATED,
	FILE_OVERWRITTEN,

	/* createoptions */
	FILE_DIRECTORY_FILE = 0x1,
	FILE_NON_DIRECTORY_FILE = 0x40,
	FILE_DELETE_ON_CLOSE = 0x1000,
	FILE_OPEN_BY_FILE_ID = 0x2000,
};
