#include <u.h>
#include <libc.h>
#include "dat.h"
#include "fns.h"

int
smberror(int class, int error)
{
	return error<<16 | (class & 0xFF);
}

int
smbmkerror(void)
{
	int err;

	static struct Ent {
		int class;
		int error;
		char *str;
	}  tab[] = {
		ERRDOS, ERRnoaccess,	"permission denied",
		ERRDOS, ERRnoaccess,	"access permission denied",
		ERRDOS, ERRnoaccess,	"create prohibited",
		ERRDOS, ERRnoaccess,	"mounted directory forbids creation",
		ERRDOS, ERRnoaccess,	"directory not empty",
		ERRDOS, ERRbadpath,	"directory entry not found",
		ERRDOS, ERRbadpath,	"not a directory",
		ERRDOS, ERRbadpath,	"does not exist",
		ERRDOS, ERRbadpath,	"bad character in directory name",
		ERRDOS, ERRbadpath,	"bad character in file name",
		ERRDOS, ERRfilexists,	"file already exists",
		ERRDOS, ERRfilexists,	"is a directory",
		/* kenfs */
		ERRDOS, ERRfilexists,	"create/wstat -- file exists",
		ERRDOS, ERRnoaccess,	"wstat -- not owner",
		ERRDOS, ERRnoaccess,	"wstat -- not in group",
		/* unknown error */
		ERRSRV, ERRerror,		nil,
	};
	char buf[ERRMAX];
	struct Ent *p;

	rerrstr(buf, sizeof(buf));
	for(p = tab; p->str; p++)
		if(strstr(buf, p->str))
			break;
	err = smberror(p->class, p->error);
	if(debug)
		fprint(2, "smbmkerror: %s -> %x\n", buf, err);
	return err;
}

int
smbmkerrorbadfile(char *path)
{
	int err;

	if((err = smbmkerror()) == smberror(ERRDOS, ERRbadpath)){
		char *x;
		Dir *d;

		x = cleanname(smprint("%s/..", path));
		d = dirstat(x);
		free(d);
		free(x);

		if(d) err = smberror(ERRDOS, ERRbadfile);

		if(debug)
			fprint(2, "smbmkerrorbadfile: %s -> %x\n", path, err);
	}
	return err;
}
