#include <u.h>
#include <libc.h>
#include "dat.h"
#include "fns.h"

File*
createfile(char *path, int dacc, int cdisp, int copt, vlong csize, int fattr, int *pact, Dir **pdir, int *perr)
{
	int err, act, fd, mode, perm, isdir, delete;
	File *f;
	Dir *d;

	f = nil;
	d = nil;
	fd = -1;

	if(copt & FILE_OPEN_BY_FILE_ID)
		goto unsup;

	mode = -1;
	if(dacc & READMASK)
		mode += 1;
	if(dacc & WRITEMASK)
		mode += 2;

	delete = isdir = 0;
	if(d = dirstat(path)){
		if(mode >= 0 && d->type != '/' && d->type != 'M')
			goto noaccess;

		isdir = d->qid.type == QTDIR;
		switch(cdisp){
		case FILE_SUPERSEDE:
			act = FILE_SUPERSEDED;
			if(remove(path) < 0){
				logit("remove: %r");
				goto oserror;
			}
			goto docreate;
		case FILE_OVERWRITE:
		case FILE_OVERWRITE_IF:
			act = FILE_OVERWRITTEN;
			if(isdir || (mode != OWRITE && mode != ORDWR))
				goto noaccess;
			d->length = 0;
			mode |= OTRUNC;
			break;
		case FILE_OPEN:
		case FILE_OPEN_IF:
			act = FILE_OPEND;
			break;
		case FILE_CREATE:
			err = smberror(ERRDOS, ERRfilexists);
			goto out;
		default:
			goto unsup;
		}
		if((copt & FILE_DIRECTORY_FILE) && !isdir)
			goto noaccess;
		if((copt & FILE_NON_DIRECTORY_FILE) && isdir)
			goto noaccess;
		if(copt & FILE_DELETE_ON_CLOSE){
			if(isdir || (dacc & FILE_DELETE)==0)
				goto noaccess;
			delete = 1;
		}
		if(mode >= 0 && !isdir)
			if((fd = open(path, mode)) < 0)
				goto oserror;
	} else {
		switch(cdisp){
		case FILE_SUPERSEDE:
		case FILE_CREATE:
		case FILE_OPEN_IF:
		case FILE_OVERWRITE_IF:
			act = FILE_CREATED;
			break;
		case FILE_OVERWRITE:
		case FILE_OPEN:
			err = smbmkerrorbadfile(path); //smberror(ERRDOS, ERRbadpath);
			goto out;
		default:
			goto unsup;
		}
docreate:
		perm = 0666;
		if(fattr & ATTR_READONLY)
			perm &= ~0222;
		if(copt & FILE_DIRECTORY_FILE){
			perm |= DMDIR | 0111;
			mode = OREAD;
			isdir = 1;
		}
		if(mode < 0)
			mode = OREAD;
		if(copt & FILE_DELETE_ON_CLOSE){
			if(isdir || (dacc & FILE_DELETE)==0)
				goto noaccess;
			delete = 1;
		}
		if((fd = create(path, mode, perm)) < 0)
			goto oserror;
		if(csize > 0 && !isdir){
			Dir nd;

			nulldir(&nd);
			nd.length = csize;
			if(dirfwstat(fd, &nd) < 0)
				goto oserror;
		}
		if(pdir)
			if((d = dirfstat(fd)) == nil)
				goto oserror;
		if(isdir){
			close(fd);
			fd = -1;
		}
	}

	err = 0;
	f = mallocz(sizeof(*f), 1);
	f->ref = 1;
	f->fd = fd; fd = -1;
	f->rtype = FileTypeDisk;
	f->path = strdup(path);
	f->access = dacc;
	f->delete = delete;

	if(pact)
		*pact = act;
	if(pdir){
		*pdir = d;
		d = nil;
	}
out:
	if(perr)
		*perr = err;
	if(fd >= 0)
		close(fd);
	free(d);
	return f;

oserror:
	err = smbmkerror();
	goto out;
noaccess:
	err = smberror(ERRDOS, ERRnoaccess);
	goto out;
unsup:
	err = smberror(ERRDOS, ERRunsup);
	goto out;
}

Dir*
statfile(File *f)
{
	if(f == nil)
		return nil;
	if(f->fd >= 0){
		return dirfstat(f->fd);
	} else {
		return dirstat(f->path);
	}
}

void
putfile(File *f)
{
	if(f == nil || --f->ref)
		return;
	if(f->fd >= 0)
		close(f->fd);
	if(f->delete){
		if(debug)
			fprint(2, "remove on close: %s\n", f->path);
		if(remove(f->path) < 0)
			logit("remove %s: %r", f->path);
	}
	free(f->path);
	free(f);
}
